/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024, 2025            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_CONVERSION_HPP_
#define PECUNIA_CONVERSION_HPP_

#include <cstdint>
#include <functional>
#include <optional>

#include <pecunia/pecunia_export.h>
#include <pecunia/Sizes.hpp>


namespace pecunia
{

enum class Currency : std::uint16_t;

}
namespace pecunia
{

class FloatingPoint;

/**
 * @brief A representation of a function that performs a rounding operation upon a real number. This
 * function must take a number, the number of digits after the whole number to round, and giving
 * back the rounded number.
 *
 * @exception pecunia::RuntimeError When the rounding is not possible.
 */
using FloatingPointRounderFunction =
		std::function<FloatingPoint(const FloatingPoint&, const std::uint8_t&)>;
/**
 * @brief A representation of a function that performs a rounding operation upon a money storage
 * type. This function must take first the major and minor amount value stored together, then the
 * number of digits in the minor unit, finally the number of digits after the whole number to round,
 * and giving back the rounded number with the major and minor amount stored together.
 *
 * @exception pecunia::RuntimeError When the rounding is not possible.
 */
using RounderFunction =
		std::function<MajorUnit(const MajorUnit&, const std::uint8_t&, const std::uint8_t)>;
/**
 * @brief A conversion function between two currencies where the first argument is the currency to
 * convert from, the second argument is the currency to convert to, and giving back the conversion
 * factor to get from the first currency to the second currency.
 *
 * @exception pecunia::RuntimeError When either the conversion from or to a currency is not
 * supported.
 */
using ConversionFunction =
		std::function<FloatingPoint(const Currency&, const Currency&)>;

/**
 * @brief Converts a truncated minor value into a minor and sub-minor value, i.e. given a value of
 * USD 25 cents it'll convert it into "2500".
 *
 * @exception pecunia::OverflowError When the supplied minor value is greater than possibly can be
 * stored in a currency's minor value.
 *
 * @param minor The minor value without the extra sub-minor amount.
 * @param code The currency the minor unit is in.
 *
 * @return The minor amount adjusted for the number of sub-minor digits and minor digits found in
 * the currency where all the sub-minor digits are padded with zeros.
 */
[[nodiscard]]
PECUNIA_EXPORT MinorUnit toMinorUnit(const MinorUnit& minor, const Currency& code);
/**
 * @brief Retrieves the exchange rate for currency conversion.
 *
 * @param from The source currency for the conversion.
 * @param to The target currency for the conversion.
 *
 * @return The exchange rate between the specified currencies if known, or null if the look-up
 * fails, or any other error occurs.
 */
[[nodiscard]]
PECUNIA_EXPORT std::optional<FloatingPoint> exchangeRate(
	const Currency& from,
	const Currency& to
) noexcept;

}

#endif
