/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024, 2025            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_ERRORS_HPP_
#define PECUNIA_ERRORS_HPP_

#include <exception>
#include <memory>
#include <string>
#include <vector>

#include <pecunia/pecunia_export.h>


namespace pecunia::internal
{

class TraceCapture;

}
namespace pecunia
{

#ifdef _WIN32
#	pragma warning(push)
// Justification: The warning "needs to have dll - interface to be used by clients of class" does
// not apply to inheritance from STL types.
#	pragma warning(disable : 4251)
// Justification: The warning "needs to have dll - interface to be used by clients of class" does
// not apply to inheritance from STL types.
#	pragma warning(disable : 4275)
#endif

/**
 * @brief A representation of a generalised software issue, complete with the location of the
 * problem.
 */
class PECUNIA_EXPORT Error : public std::exception
{
	std::string message_;
	std::unique_ptr<internal::TraceCapture> trace_;

public:
	/**
 	 * Default constructor that initialises with a default message and captures the stack trace.
	 */
	Error() noexcept;
	/**
	 * @brief Message constructor that initialises with a supplied message and captures the
	 * backtrace for diagnostic purposes.
	 *
	 * @param message The details providing context about the error to be described in the instance.
	 */
	explicit Error(std::string message) noexcept;
	/**
	 * @brief Constructs an Error object with a synthesised message combining the provided message,
	 * and the error message output of the supplied original exception.
	 *
	 * @param message The details providing context about the error to be described in the instance.
	 * @param original The original exception that caused this Error, used to extract additional
	 * information.
	 */
	Error(const std::string& message, const std::exception& original) noexcept;
	/**
	 * @brief Constructs a copy of an existing instance.
	 *
	 * It performs a deep copy of the internals, ensuring uniqueness of resources between the
	 * original and the copy.
	 *
	 * @param other The other instance to copy from.
	 *
	 * @post The newly constructed object will have the same message and trace information as the
	 * original `other` instance.
	 */
	Error(const Error& other) noexcept;
	/**
	 * @brief Constructs an object by transferring ownership of resources from another object.
	 *
	 * This constructor is used to create a new instance, taking over the existing data and
	 * state of another object.
	 *
	 * @param other The other object whose resources will be moved to the newly created object.
	 *
	 * @post The newly created object contains the previous state of the `other` object. The `other`
	 * object is left in a valid, but unspecified state.
	 */
	Error(Error&& other) noexcept;
	/**
	 * @brief Safely destroys the object, ensuring any allocated resources are properly freed.
	 */
	~Error() noexcept override;
	/**
	 * @brief Assigns the state of the given instance to this instance.
	 *
	 * This method ensures that all internal states of the current object are replaced by the
	 * corresponding states from the `other` object. The method aims to establish functional
	 * equality between the current instance, and the `other` instance.
	 *
	 * @param other The instance whose state will be copied to the current object.
	 *
	 * @post The current instance will contain the same state as the `other` instance.
	 *
	 * @return A reference to the current instance, reflecting the updated state.
	 */
	Error& operator=(const Error& other) noexcept;
	/**
	 * @brief Assigns the given object to the current instance, transferring ownership of resources
	 * if any.
	 *
	 * @param other The object that will be assigned to the current instance.
	 *
	 * @post The current instance will contain the data or configuration of the other object.
	 * @post The “other” object will be left in a valid but unspecified state.
	 *
	 * @return The current instance after the assignment.
	 */
	Error& operator=(Error&& other) noexcept;
	/**
	 * @brief Retrieves the human-readable description of the error. It is commonly used for logging
	 * or display purposes.
	 *
	 * @return The error’s message, which may provide insight into what caused the error.
	 */
	[[nodiscard]]
	const char* what() const noexcept override;
	/**
	 * @brief Retrieves the stack frames associated with this instance.
	 *
	 * This is used to obtain a backtrace of stack frames that can help in identifying where
	 * the error occurred. It generates a detailed trace based on the error’s internal state.
	 *
	 * @return A vector of strings representing the sequence of stack frames of the error’s
	 * backtrace. The specific formatting is dependent upon the compiler, but the error’s
	 * constructor will /not/ be a part of the trace.
	 *
	 * @note The trace is generated on demand, and is not cached.
	 */
	[[nodiscard]]
	std::vector<std::string> frameStack() const noexcept;
};

#ifdef _WIN32
#	pragma warning(pop)
#endif

/**
 * @brief Represents a flaw in the logic or a precondition of a function.
 */
struct PECUNIA_EXPORT LogicError : Error
{
	using Error::Error;
};
/**
 * @brief Represents an argument that is invalid for the state of the function.
 */
struct PECUNIA_EXPORT InvalidArgument : LogicError
{
	using LogicError::LogicError;
};
/**
 * @brief Represents an argument that is outside the realm of the purpose of a function.
 */
struct PECUNIA_EXPORT DomainError : LogicError
{
	using LogicError::LogicError;
};
/**
 * @brief Represents an argument that is out of range for a container.
 */
struct PECUNIA_EXPORT OutOfRange : LogicError
{
	using LogicError::LogicError;
};
/**
 * @brief Represents a flaw that prevents a function from fulfilling its postcondition or purpose.
 */
struct PECUNIA_EXPORT RuntimeError : Error
{
	using Error::Error;
};
/**
 * @brief Represents a flaw that would trigger an overflow operation.
 */
struct PECUNIA_EXPORT OverflowError : RuntimeError
{
	using RuntimeError::RuntimeError;
};
/**
 * @brief Represents a flaw that would trigger an underflow operation.
 */
struct PECUNIA_EXPORT UnderflowError : RuntimeError
{
	using RuntimeError::RuntimeError;
};

}

#endif
