/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024, 2025            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_EXPECTED_HPP_
#define PECUNIA_EXPECTED_HPP_

#include <cstddef>
#include <functional>
#include <type_traits>
#include <variant>

#include <pecunia/pecunia_export.h>


namespace pecunia
{

/**
 * @brief Represents a globally accessible, inline constant placeholder for an unexpected error.
 * It is used to handle or signify error states in a consistent and type-safe manner.
 */
static struct PECUNIA_EXPORT
{} inline constexpr unexpectedErr;

/**
 * @brief Represents an unexpected error value.
 *
 * It encapsulates an error state when a computation or operation does not yield a valid value but
 * instead produces an error. It is primarily used to signify and propagate error states in
 * expected/optional-like patterns in an explicit and type-safe manner.
 *
 * @tparam ErrorType The type used to represent the error value.
 */
template<typename ErrorType>
struct Unexpected
{
	/**
	 * @brief Full initialisation constructor.
	 *
	 * @param value The error value to encapsulate.
	 */
	explicit Unexpected(ErrorType value);

	/// Represents the underlying error value for an operation that failed or holds an invalid
	/// state.
	ErrorType value_;
};
/**
 * @brief Represents a result value that may either contain a valid value or an error state. Useful
 * for error handling and propagating error states alongside valid results. Provides a way to work
 * with results without using exceptions or nullable types. Ensures type safety by enforcing the
 * presence of either a value or an error.
 *
 * @pre A correctly initialised object must contain either a value or an error, but not both.
 * @pre Neither the value nor the error should be in an uninitialised or invalid state.
 *
 * @tparam ValueType The type of the success value that can be encapsulated in the object.
 * @tparam ErrorType The type representing an error state for the object, defaulting to bool.
 */
template<typename ValueType, typename ErrorType = bool>
class Expected
{
	static_assert(
		 ! std::is_reference_v<ValueType>,
		"Cannot have an expected value type that is a reference."
	);
	static_assert(
		 ! std::is_same_v<ValueType, std::in_place_t>,
		"Cannot have an expected value type that is the std::in_place_t type."
	);
	static_assert(
		 ! std::is_same_v<ValueType, decltype(unexpectedErr)>,
		"Cannot have an expected value type that is the unexpected error type."
	);
	static_assert(
		 ! std::is_reference_v<ErrorType>,
		"Cannot have an expected error type that is a reference."
	);
	static_assert(
		 ! std::is_invocable_v<ValueType>,
		"Cannot have an expected value type that is an invocable."
	);
	static_assert(
		 ! std::is_invocable_v<ErrorType>,
		"Cannot have an expected error type that is an invocable."
	);
	static_assert(
		std::is_destructible_v<ValueType>,
		"Cannot have an expected value type that is not destructible."
	);
	static_assert(
		std::is_destructible_v<ErrorType>,
		"Cannot have an expected error type that is not destructible."
	);

	static constexpr std::size_t valueIndex{0};
	static constexpr std::size_t errorIndex{1};
	
	std::variant<ValueType, ErrorType> valOrErr_;

public:
	/**
	 * @brief Default constructor that initialises the object to a state with no value present and
	 * a default-constructed error object. The default error value depends on the error type's
	 * default constructor.
	 */
	Expected();
	/**
	 * @brief Constructs an instance with a success value.
	 *
	 * @tparam Parameters Variadic template representing the types of parameters for ValueType's
	 * constructor.
	 *
	 * @param params The parameters to be forwarded to the constructor of ValueType.
	 */
	template<typename ... Parameters>
	explicit Expected(std::in_place_t, Parameters&& ... params);
	/**
	 * @brief Constructs an instance with an error value.
	 *
	 * @tparam Parameters Variadic template representing the types of parameters for ErrorType's
	 * constructor.
	 *
	 * @param params The parameters to be forwarded to the constructor of ErrorType.
	 */
	template<typename ... Parameters>
	explicit Expected(decltype(unexpectedErr), Parameters&& ... params);
	/**
	 * @brief Assigns a new success value.
	 *
	 * @param value The value to be assigned to the object.
	 *
	 * @post The object will hold the newly assigned value, replacing any previously stored value.
	 * @post The object will no longer contain an error state.
	 *
	 * @return A reference to the modified object containing the newly assigned value.
	 */
	Expected& operator=(const ValueType& value);
	/**
	 * @brief Assigns a new success value.
	 *
	 * @param value The value to be assigned to the object.
	 *
	 * @post The object will hold the newly assigned value, replacing any previously stored value.
	 * @post The object will no longer contain an error state.
	 *
	 * @return A reference to the modified object containing the newly assigned value.
	 */
	Expected& operator=(ValueType&& value);
	/**
	 * @brief Assigns an unexpected error value to the current instance. This is only available when
	 * `ValueType` and `ErrorType` are different.
	 *
	 * @tparam ErrType The error type, this must be the same as `ErrorType`.
	 *
	 * @param error The object containing the error value to be assigned.
	 *
	 * @post The instance will contain an error state with the specified error value. Any previously
	 * held value or error will be replaced.
	 *
	 * @return A reference to the updated instance after the assignment.
	 */
	template<typename ErrType>
	Expected& operator=(const Unexpected<ErrType>& error);
	/**
	 * @brief Assigns an unexpected error value to the current instance. This is only available when
	 * `ValueType` and `ErrorType` are different.
	 *
	 * @tparam ErrType The error type, this must be the same as `ErrorType`.
	 *
	 * @param error The object containing the error value to be assigned.
	 *
	 * @post The instance will contain an error state with the specified error value. Any previously
	 * held value or error will be replaced.
	 *
	 * @return A reference to the updated instance after the assignment.
	 */
	template<typename ErrType>
	Expected& operator=(Unexpected<ErrType>&& error);
	/**
	 * @brief Assigns an unexpected error value to the current instance. This is only available when
	 * `ValueType` and `ErrorType` are different.
	 *
	 * @tparam ErrType The error type, this must be the same as `ErrorType`.
	 *
	 * @param error The object containing the error value to be assigned.
	 *
	 * @post The instance will contain an error state with the specified error value. Any previously
	 * held value or error will be replaced.
	 *
	 * @return A reference to the updated instance after the assignment.
	 */
	template<typename ErrType, typename = std::enable_if_t< ! std::is_same_v<ValueType, ErrType>>>
	Expected& operator=(const ErrType& error)
	{
		this->valOrErr_.template emplace<Expected::errorIndex>(error);
		return *this;
	}
	/**
	 * @brief Assigns an unexpected error value to the current instance. This is only available when
	 * `ValueType` and `ErrorType` are different.
	 *
	 * @tparam ErrType The error type, this must be the same as `ErrorType`.
	 *
	 * @param error The object containing the error value to be assigned.
	 *
	 * @post The instance will contain an error state with the specified error value. Any previously
	 * held value or error will be replaced.
	 *
	 * @return A reference to the updated instance after the assignment.
	 */
	template<typename ErrType, typename = std::enable_if_t< ! std::is_same_v<ValueType, ErrType>>>
	Expected& operator=(ErrType&& error)
	{
		this->valOrErr_.template emplace<Expected::errorIndex>(std::move(error));
		return *this;
	}
	/**
	 * @brief Determines whether the object contains a valid value.
	 *
	 * This method is used to check if the object currently holds a valid value (indicating
	 * success), as opposed to an error state. It can be used to verify the success of operations
	 * or when a value is required for subsequent processing.
	 *
	 * @return Gives true if the object contains a value; false if it contains an error.
	 */
	[[nodiscard]]
	bool hasValue() const noexcept;
	/**
	 * @brief Provides pointer-like access to the stored value if the operation is successful.
	 *
	 * @pre The instance contains a valid value. This must be ensured prior to invoking the method;
	 * otherwise, the behaviour is undefined.
	 *
	 * @return A pointer to the contained value if the object holds a value.
	 */
	[[nodiscard]]
	const ValueType* operator->() const;
	/**
	 * @brief Provides access to the contained value if no error is present.
	 *
	 * @pre The instance contains a valid value. This must be ensured prior to invoking the method;
	 * otherwise, the behaviour is undefined.
	 *
	 * @return A pointer to the contained value if the object holds a value.
	 */
	[[nodiscard]]
	ValueType* operator->();
	/**
	 * @brief Provides access to the stored value within the instance if it contains a value.
	 *
	 * It is primarily used when you are certain the instance holds a value, and you intend to
	 * operate on or retrieve that value.
	 *
	 * @pre The instance contains a valid value. This must be ensured prior to invoking the method;
	 * otherwise, the behaviour is undefined.
	 *
	 * @return The stored value of the type.
	 */
	[[nodiscard]]
	const ValueType& operator*() const &;
	/**
	 * @brief Provides access to the stored value within the instance if it contains a value.
	 *
	 * It is primarily used when you are certain the instance holds a value, and you intend to
	 * operate on or retrieve that value.
	 *
	 * @pre The instance contains a valid value. This must be ensured prior to invoking the method;
	 * otherwise, the behaviour is undefined.
	 *
	 * @return The stored value of the type.
	 */
	[[nodiscard]]
	ValueType& operator*() &;
	/**
	 * @brief Provides access to the stored value within the instance if it contains a value.
	 *
	 * It is primarily used when you are certain the instance holds a value, and you intend to
	 * operate on or retrieve that value.
	 *
	 * @pre The instance contains a valid value. This must be ensured prior to invoking the method;
	 * otherwise, the behaviour is undefined.
	 *
	 * @return The stored value of the type.
	 */
	[[nodiscard]]
	ValueType&& operator*() &&;
	/**
	 * @brief Provides access to the stored value within the instance if it contains a value.
	 *
	 * It is primarily used when you are certain the instance holds a value, and you intend to
	 * operate on or retrieve that value.
	 *
	 * @pre The instance contains a valid value. This must be ensured prior to invoking the method;
	 * otherwise, the behaviour is undefined.
	 *
	 * @return The stored value of the type.
	 */
	[[nodiscard]]
	const ValueType& value() const &;
	/**
	 * @brief Provides access to the stored value within the instance if it contains a value.
	 *
	 * It is primarily used when you are certain that the instance holds a value, and you intend to
	 * operate on or retrieve that value.
	 *
	 * @pre The instance contains a valid value. This must be ensured prior to invoking the method;
	 * otherwise, the behaviour is undefined.
	 *
	 * @return The stored value of the type.
	 */
	[[nodiscard]]
	ValueType& value() &;
	/**
	 * @brief Provides access to the stored value within the instance if it contains a value.
	 *
	 * It is primarily used when you are certain that the instance holds a value, and you intend to
	 * operate on or retrieve that value.
	 *
	 * @pre The instance contains a valid value. This must be ensured prior to invoking the method;
	 * otherwise, the behaviour is undefined.
	 *
	 * @return The stored value of the type.
	 */
	[[nodiscard]]
	ValueType&& value() &&;
	template<typename ElseValueType>
	/**
	 * @brief Retrieves the contained value if it exists; otherwise, returns the provided
	 * alternative value.
	 *
	 * @tparam ElseValueType The type of the alternative value to be returned if no valid value
	 * exists. It must be the same as `ValueType`.
	 *
	 * @param elseValue The alternative value to return if the instance does not contain a valid
	 * value.
	 *
	 * @return The contained value if it exists, or the provided alternative value if no valid value
	 * is present.
	 */
	[[nodiscard]]
	ValueType valueOr(ElseValueType&& elseValue) const;
	/**
	 * @brief Retrieves the stored error value from the object if it contains an error.
	 *
	 * This function is useful for accessing the error value from an object, allowing the handling
	 * of error cases when the object does not contain a valid value. It is intended to be called
	 * on objects where the error state is known to be present.
	 *
	 * @pre The Expected object must be in an error state. If the object is not in an error state,
	 * calling this function results in undefined behaviour.
	 *
	 * @return The stored error value.
	 */
	[[nodiscard]]
	const ErrorType& error() const &;
	/**
	 * @brief Retrieves the stored error value from the object if it contains an error.
	 *
	 * This function is useful for accessing the error value from an object, allowing the handling
	 * of error cases when the object does not contain a valid value. It is intended to be called
	 * on objects where the error state is known to be present.
	 *
	 * @pre The Expected object must be in an error state. If the object is not in an error state,
	 * calling this function results in undefined behaviour.
	 *
	 * @return The stored error value.
	 */
	[[nodiscard]]
	ErrorType& error() &;
	/**
	 * @brief Retrieves the stored error value from the object if it contains an error.
	 *
	 * This function is useful for accessing the error value from an object, allowing the handling
	 * of error cases when the object does not contain a valid value. It is intended to be called
	 * on objects where the error state is known to be present.
	 *
	 * @pre The Expected object must be in an error state. If the object is not in an error state,
	 * calling this function results in undefined behaviour.
	 *
	 * @return The stored error value.
	 */
	[[nodiscard]]
	ErrorType&& error() &&;
};

/**
 * @brief Applies a specified operation based on the state of the provided object.
 *
 * This method invokes the appropriate function depending on whether the object contains a value or
 * an error. It is useful for handling objects in a structured and functional way, ensuring the
 * appropriate operation is applied based on the internal state of the object.
 *
 * @pre The provided functions must be callable with their respective types.
 *
 * @tparam ValueType The type wrapped as the success value inside the object.
 * @tparam ErrorType The type wrapped as the error inside the object.
 *
 * @param expected The object to process. It contains either ValueType or ErrorType.
 * @param matchesValueType The function to be applied if the object contains a success value.
 * @param matchesErrorType The function to be applied if the object contains an error.
 *
 * @post The selected function is applied to the expected’s internal success value or internal
 * error respectively.
 *
 * @return The modified object after applying the appropriate function.
 */
template<typename ValueType, typename ErrorType>
Expected<ValueType, ErrorType>& apply(
	Expected<ValueType, ErrorType>& expected,
	const std::function<void(ValueType&)>& matchesValueType,
	const std::function<void(ErrorType&)>& matchesErrorType
);
/**
 * @brief Applies a specified operation based on the state of the provided object.
 *
 * This method invokes the appropriate function depending on whether the object contains a value or
 * an error. It is useful for handling objects in a structured and functional way, ensuring the
 * appropriate operation is applied based on the internal state of the object.
 *
 * @pre The provided functions must be callable with their respective types.
 *
 * @tparam ValueType The type wrapped as the success value inside the object.
 * @tparam ErrorType The type wrapped as the error inside the object.
 *
 * @param expected The object to process. It contains either ValueType or ErrorType.
 * @param matchesValueType The function to be applied if the object contains a success value.
 * @param matchesErrorType The function to be applied if the object contains an error.
 *
 * @return The unmodified object.
 */
template<typename ValueType, typename ErrorType>
const Expected<ValueType, ErrorType>& apply(
	const Expected<ValueType, ErrorType>& expected,
	const std::function<void(const ValueType&)>& matchesValueType,
	const std::function<void(const ErrorType&)>& matchesErrorType
);

}

#include "Expected.tpp"

#endif
