/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024, 2025            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_EXPECTED_HPP_
#error Template definiton files should never be included directly.
#endif

#include <cassert>
#include <utility>


template<typename ErrorType>
pecunia::Unexpected<ErrorType>::Unexpected(ErrorType value) :
	value_{value}
{}

template<typename ValueType, typename ErrorType>
pecunia::Expected<ValueType, ErrorType>::Expected():
	valOrErr_{std::in_place_index<Expected::errorIndex>, ErrorType{}}
{}

template<typename ValueType, typename ErrorType>
template<typename ... Parameters>
pecunia::Expected<ValueType, ErrorType>::Expected(std::in_place_t, Parameters&& ... params) :
	valOrErr_{std::in_place_index<Expected::valueIndex>, std::forward<Parameters>(params)...}
{}

template<typename ValueType, typename ErrorType>
template<typename ... Parameters>
pecunia::Expected<ValueType, ErrorType>::Expected(
	decltype(unexpectedErr),
	Parameters&& ... params
) :
	valOrErr_{std::in_place_index<Expected::errorIndex>, std::forward<Parameters>(params)...}
{}

template<typename ValueType, typename ErrorType>
pecunia::Expected<ValueType, ErrorType>& pecunia::Expected<ValueType, ErrorType>::operator=(
	const ValueType& value
)
{
	this->valOrErr_.template emplace<Expected::valueIndex>(value);
	return *this;
}

template<typename ValueType, typename ErrorType>
pecunia::Expected<ValueType, ErrorType>& pecunia::Expected<ValueType, ErrorType>::operator=(
	ValueType&& value
)
{
	this->valOrErr_.template emplace<Expected::valueIndex>(std::move(value));
	return *this;
}

template<typename ValueType, typename ErrorType>
template<typename ErrType>
pecunia::Expected<ValueType, ErrorType>& pecunia::Expected<ValueType, ErrorType>::operator=(
	const pecunia::Unexpected<ErrType>& error
)
{
	this->valOrErr_.template emplace<Expected::errorIndex>(error.value_);
	return *this;
}

template<typename ValueType, typename ErrorType>
template<typename ErrType>
pecunia::Expected<ValueType, ErrorType>& pecunia::Expected<ValueType, ErrorType>::operator=(
	pecunia::Unexpected<ErrType>&& error
)
{
	this->valOrErr_.template emplace<Expected::errorIndex>(std::move(error.value_));
	return *this;
}

template<typename ValueType, typename ErrorType>
bool pecunia::Expected<ValueType, ErrorType>::hasValue() const noexcept
{
	return this->valOrErr_.index() == Expected::valueIndex;
}

template<typename ValueType, typename ErrorType>
const ValueType* pecunia::Expected<ValueType, ErrorType>::operator->() const
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return &std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ValueType* pecunia::Expected<ValueType, ErrorType>::operator->()
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return &std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
const ValueType& pecunia::Expected<ValueType, ErrorType>::operator*() const &
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ValueType& pecunia::Expected<ValueType, ErrorType>::operator*() &
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ValueType&& pecunia::Expected<ValueType, ErrorType>::operator*() &&
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::move(std::get<Expected::valueIndex>(this->valOrErr_));
}

template<typename ValueType, typename ErrorType>
const ValueType& pecunia::Expected<ValueType, ErrorType>::value() const &
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ValueType& pecunia::Expected<ValueType, ErrorType>::value() &
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::get<Expected::valueIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ValueType&& pecunia::Expected<ValueType, ErrorType>::value() &&
{
	assert(this->hasValue() && "Accessing an expected value must always have a value.");
	return std::move(std::get<Expected::valueIndex>(this->valOrErr_));
}

template<typename ValueType, typename ErrorType>
template<typename ElseValueType>
ValueType pecunia::Expected<ValueType, ErrorType>::valueOr(ElseValueType&& elseValue) const
{
	static_assert(
		std::is_convertible_v<ValueType, ElseValueType>,
		"The default value type must be a convertable type to the expected type."
	);

	if (this->hasValue())
		return this->value();
	return elseValue;
}

template<typename ValueType, typename ErrorType>
const ErrorType& pecunia::Expected<ValueType, ErrorType>::error() const &
{
	assert( ! this->hasValue() && "Accessing an expected error must never have a value.");
	return std::get<Expected::errorIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ErrorType& pecunia::Expected<ValueType, ErrorType>::error() &
{
	assert( ! this->hasValue() && "Accessing an expected error must never have a value.");
	return std::get<Expected::errorIndex>(this->valOrErr_);
}

template<typename ValueType, typename ErrorType>
ErrorType&& pecunia::Expected<ValueType, ErrorType>::error() &&
{
	assert( ! this->hasValue() && "Accessing an expected error must never have a value.");
	return std::move(std::get<Expected::errorIndex>(this->valOrErr_));
}

template<typename ValueType, typename ErrorType>
pecunia::Expected<ValueType, ErrorType>& pecunia::apply(
	pecunia::Expected<ValueType, ErrorType>& expected,
	const std::function<void(ValueType&)>& matchesValueType,
	const std::function<void(ErrorType&)>& matchesErrorType
)
{
	assert(matchesValueType != nullptr && "The function for a matching value cannot be null.");
	assert(matchesErrorType != nullptr && "The function for a matching error cannot be null.");

	if (expected.hasValue())
		matchesValueType(*expected);
	else
		matchesErrorType(expected.error());
	return expected;
}

template<typename ValueType, typename ErrorType>
const pecunia::Expected<ValueType, ErrorType>& pecunia::apply(
	const pecunia::Expected<ValueType, ErrorType>& expected,
	const std::function<void(const ValueType&)>& matchesValueType,
	const std::function<void(const ErrorType&)>& matchesErrorType
)
{
	assert(matchesValueType != nullptr && "The function for a matching value cannot be null.");
	assert(matchesErrorType != nullptr && "The function for a matching error cannot be null.");

	if (expected.hasValue())
		matchesValueType(*expected);
	else
		matchesErrorType(expected.error());
	return expected;
}
