/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024, 2025            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_MANIP_HPP_
#define PECUNIA_MANIP_HPP_

#include <iomanip>

#include <pecunia/pecunia_export.h>


namespace pecunia
{

/**
 * @brief A stream manipulator that ensures the currency indicator is displayed before the value in
 * formatted output.
 *
 * @param stream The stream to apply the manipulator to.
 *
 * @post The stream’s internal state is updated to position the currency indicator before the value.
 *
 * @return The manipulated stream.
 */
[[nodiscard]]
PECUNIA_EXPORT std::ios_base& leading(std::ios_base& stream);
/**
 * @brief A stream manipulator that positions the currency indicator after the monetary value in
 * formatted output.
 *
 * @param stream The stream to apply the manipulator to.
 *
 * @post The stream’s internal state is updated to position the currency indicator after the value.
 *
 * @return The manipulated stream.
 */
[[nodiscard]]
PECUNIA_EXPORT std::ios_base& trailing(std::ios_base& stream);
/**
 * @brief Sets the stream to apply a currency symbol formatting behaviour during output operations.
 *
 * @param stream The stream on which to set the currency symbol formatting behaviour.
 *
 * @post The stream’s internal state is updated to enable the display of the currency symbol.
 *
 * @return The modified stream with the currency symbol formatting behaviour enabled.
 */
[[nodiscard]]
PECUNIA_EXPORT std::ios_base& symbol(std::ios_base& stream);
/**
 * @brief A stream manipulator that configures the currency representation in formatted output to
 * use the ISO-4217 code instead of the currency symbol.
 *
 * @param stream The stream to apply the manipulator to.
 *
 * @post The stream’s internal state is updated to reflect the use of the ISO-4217 code for currency
 * representation in place of the symbol.
 *
 * @return The manipulated stream.
 */
[[nodiscard]]
PECUNIA_EXPORT std::ios_base& isocode(std::ios_base& stream);
/**
 * @brief A stream manipulator that ensures a space is inserted appropriately between components
 * in formatted monetary output.
 *
 * @param stream The stream to apply the manipulator to.
 *
 * @post The stream’s internal state is updated to indicate that spacing should be applied
 * between monetary components.
 *
 * @return The manipulated stream.
 */
[[nodiscard]]
PECUNIA_EXPORT std::ios_base& spaced(std::ios_base& stream);
/**
 * @brief A stream manipulator that disables spacing between the currency symbol, and the numeric
 * value in formatted output.
 *
 * @param stream The stream to apply the manipulator to.
 *
 * @post The stream’s internal state is updated to enforce no spaces between the currency symbol
 * and the numeric value.
 *
 * @return The manipulated stream.
 */
[[nodiscard]]
PECUNIA_EXPORT std::ios_base& nonspaced(std::ios_base& stream);

}

#endif
