/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024, 2025            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_MATH_HPP_
#define PECUNIA_MATH_HPP_

#include <functional>
#include <optional>

#include <pecunia/pecunia_export.h>

#include "Currency.hpp"
#include "Money.hpp"


namespace pecunia::math
{

/**
 * @brief Calculates the absolute value for the supplied amount of money.
 *
 * @param money The money whose monetary amount should be an absolute value.
 *
 * @return The absolute value of the monetary amount.
 */
[[nodiscard]]
PECUNIA_EXPORT Money absoluteValue(const Money& money);
/**
 * @brief Sums up a collection of monetary values, potentially transformed by a user-defined mutator
 * function, with the fewest possible number of conversions.
 *
 * @pre When a mutator is supplied, it cannot be null.
 * @pre When a container’s element is not a Money type, the adjustor must be supplied.
 *
 * @tparam Container A container type that holds values convertible to pecunia::Money.
 *
 * @param accumulateIn The currency in which the result will be accumulated.
 * @param elements The container of elements to be summed.
 * @param mutator An optional function to transform container elements to pecunia::Money
 * before summing or apply an adjustment before summing.
 *
 * @return The sum of all elements in the container in the currency provided.
 */
template<typename Container>
[[nodiscard]]
Money sum(
	const Currency& accumulateIn,
	const Container& elements,
	const std::optional<
		std::function<Money(const typename Container::value_type&)>
	>& mutator = std::nullopt
);
/**
 * @brief Accumulates the total of all the supplied monies with the fewest possible number of
 * conversions.
 *
 * @pre When an adjustor is supplied, it cannot be null.
 *
 * @tparam AssociativeContainer The associative container type holding the money.
 * @tparam KeyType The type the monies are keyed against, not used.
 * @tparam ValueType The value type in the container, which must include monetary amounts.
 * @tparam Compare The comparison type used for ordering elements in the container.
 * @tparam Allocator The allocator type used for memory management in the container.
 *
 * @param accumulateIn The currency in which the final sum will be accumulated.
 * @param elements The container holding monetary pairs to be summed.
 * @param adjustor An optional function to adjust each monetary value during summation. If not
 * provided, no adjustments are applied.
 *
 * @return The result of the summation as a monetary value in the specified currency.
 */
template<
	template<typename, typename, typename, typename>
	typename AssociativeContainer,
	typename KeyType,
	typename ValueType,
	typename Compare = std::less<KeyType>,
	typename Allocator = std::allocator<std::pair<const KeyType, ValueType>>
>
[[nodiscard]]
Money sum(
	const Currency& accumulateIn,
	const AssociativeContainer<KeyType, ValueType, Compare, Allocator>& elements,
	const std::optional<
		std::function<Money(const Money& money)>
	>& adjustor = std::nullopt
);

}

#include "Math.tpp"

#endif
