/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024, 2025            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_MATH_HPP_
#error Template definiton files should never be included directly.
#endif

#include <cassert>
#include <map>
#include <type_traits>


namespace pecunia::math::detail
{

[[nodiscard]]
PECUNIA_EXPORT Money accumulateRunningTotals(
	const Currency& accumulateCurrency,
	const std::map<Currency, Money>& totals
);

}

template<typename Container>
pecunia::Money pecunia::math::sum(
	const pecunia::Currency& accumulateIn,
	const Container& elements,
	const std::optional<
		std::function<Money(const typename Container::value_type&)>
	>& mutator
)
{
	assert(
		( ! mutator.has_value() || *mutator != nullptr)
		&& "When an adjustor is supplied, it cannot be null."
	);
	std::map<pecunia::Currency, pecunia::Money> runningTotals{};

	for (const auto& element : elements)
	{
		const pecunia::Money amount{
			[&mutator, &element] ()
			{
				if constexpr (std::is_same_v<typename Container::value_type, Money>)
				{
					return mutator.has_value() ? (*mutator)(element) : element;
				}
				else
				{
					assert(
						mutator.has_value()
						&& "The mutator can't be null when the container's element is not a Money "
							"type."
					);
					return (*mutator)(element);
				}
			}()
		};

		if (runningTotals.count(amount.code()) == 0)
			runningTotals[amount.code()].assign(amount);
		else
			runningTotals[amount.code()] += amount;
	}
	return detail::accumulateRunningTotals(accumulateIn, runningTotals);
}

template<
	template <typename, typename, typename, typename>
	class AssociativeContainer,
	class KeyType,
	class ValueType,
	class Compare,
	class Allocator
>
pecunia::Money pecunia::math::sum(
	const Currency& accumulateIn,
	const AssociativeContainer<KeyType, ValueType, Compare, Allocator>& elements,
	const std::optional<std::function<Money(const Money& money)>>& adjustor
)
{
	assert(
		( ! adjustor.has_value() || *adjustor != nullptr)
		&& "When an adjustor is supplied, it cannot be null."
	);
	static_assert(
		std::is_same_v<ValueType, Money>,
		"The container's value type must be a Money type."
	);
	std::map<pecunia::Currency, pecunia::Money> runningTotals{};

	for (const auto& element : elements)
	{
		const Money amount{
			adjustor.has_value() ? (*adjustor)(element.second) : element.second
		};

		if (runningTotals.count(amount.code()) == 0)
			runningTotals[amount.code()].assign(amount);
		else
			runningTotals[amount.code()] += amount;
	}
	return detail::accumulateRunningTotals(accumulateIn, runningTotals);
}
