/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024, 2025            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_MONEY_HPP_
#define PECUNIA_MONEY_HPP_

#include <cstdint>
#include <iosfwd>
#include <optional>
#include <string>

#include <pecunia/Sizes.hpp>

#include "Currency.hpp"
#include "Conversion.hpp"


namespace pecunia
{

class FloatingPoint;
class Money;
class MoneyParts;

}
namespace pecunia::math
{

PECUNIA_EXPORT Money absoluteValue(const Money&);

}
namespace pecunia
{

/**
 * @brief Represents a monetary value in minor units, with associated currency code and
 * sign/affinity.
 */
class PECUNIA_EXPORT Minor
{
public:
	/**
	 * @brief When working only with minor units, this represents the positivity or negativity of a
	 * minor value.
	 */
	enum class PECUNIA_EXPORT Sign : std::uint8_t
	{
		/// The minor value is negative.
		Negative,
		/// The minor value is neither positive nor negative, i.e. zero.
		Neither,
		/// The minor value is positive.
		Positive,
	};

private:
	Sign affinity_;
	Currency code_;
	MinorUnit amount_;

public:
	/**
	 * Default unknown currency constructor.
	 */
	Minor();
	/**
	 * @brief Sub-minor initialisation constructor. This is largely useful for currencies that do
	 * not support minor currencies and need to have a negative sub-minor value.
	 *
	 * @param subMinor Specifies the value for the sub-minor portion of the minor unit.
	 * @param code Represents the currency code associated with the monetary value.
	 */
	Minor(const MajorUnit subMinor, const Currency& code);
	/**
	 * @brief Full initialisation constructor for a currency that requires a minor value.
	 *
	 * @exception InvalidArgument Thrown if the `minor` value is provided while the currency does
	 * not support minor digits.
	 * @exception InvalidArgument Thrown if the `minor` value is /not/ provided while the currency
	 * *does* support minor digits.
	 * @exception OverflowError Thrown if the `minor` value exceeds the maximum allowable value for
	 * the given currency’s minor digits.
	 * @exception OverflowError Thrown if the `subMinor` value exceeds the maximum allowable
	 * sub-minor value.
	 *
	 * @param code The currency associated with this instance.
	 * @param minor A value representing the minor unit of the monetary value.
	 * @param subMinor A value representing the sub-minor unit of the monetary value.
	 */
	explicit Minor(
		const Currency& code,
		const MajorUnit minor = 0u,
		const MinorUnit subMinor = 0u
	);
	/**
	 * @brief Full initialisation constructor representing a monetary value in its associated
	 * currency with minor (when supported) and sub-minor units.
	 *
	 * @exception InvalidArgument Thrown if the `minor` value is provided while the currency does
	 * not support minor digits.
	 * @exception InvalidArgument Thrown if the `minor` value is /not/ provided while the currency
	 * *does* support minor digits.
	 * @exception OverflowError Thrown if the `minor` value exceeds the maximum allowable value for
	 * the given currency’s minor digits.
	 * @exception OverflowError Thrown if the `subMinor` value exceeds the maximum allowable
	 * sub-minor value.
	 *
	 * @param code The currency associated with this instance.
	 * @param minor An optional value representing the minor unit of the monetary value.
	 * @param subMinor A value representing the sub-minor unit of the monetary value.
	 */
	Minor(
		const Currency& code,
		const std::optional<MajorUnit> minor,
		const MinorUnit subMinor = 0u
	);
	/**
	 * @brief An accessor to the sign/affinity of the minor currency.
	 *
	 * @return The affinity/sign of the object.
	 */
	[[nodiscard]]
	Sign affinity() const noexcept;
	/**
	 * @brief Retrieves the currency code associated with the monetary value.
	 *
	 * @return The currency code of the monetary value’s currency.
	 */
	[[nodiscard]]
	Currency code() const noexcept;
	/**
	 * @brief Retrieves the minor unit representation of the monetary value, if available.

	 * @return When the currency supports minor units, the minor unit value, or an empty value when
	 * it does not.
	 */
	[[nodiscard]]
	std::optional<MinorUnit> minorOnly() const noexcept;
	/**
	 * @brief Computes the total monetary value in minor units, including any sub-minor
	 * adjustments, based on the stored amount, currency code, and affinity (sign).

	 * @return The value representing the full monetary amount in minor units, adjusted for
	 * sub-minor details and the sign affinity.
	 */
	[[nodiscard]]
	MajorUnit minorWithSub() const noexcept;
	/**
	 * @brief Returns the sub-minor unit representation of the monetary value.
	 * This value represents finer resolution units beneath the primary minor unit of currency.
	 *
	 * @return The sub-minor unit equivalent of the monetary value.
	 */
	[[nodiscard]]
	MinorUnit subMinor() const noexcept;
};
/**
 * @brief A representation of a country’s currency supporting basic arithmetic usage and
 * guaranteeing all operations succeed without a loss of precision, overflow, or underflow.
 */
class PECUNIA_EXPORT Money
{
	friend PECUNIA_EXPORT Money math::absoluteValue(const Money&);
	friend PECUNIA_EXPORT Money operator*(const std::int8_t& lhs, const Money& rhs);
	friend PECUNIA_EXPORT Money operator*(const std::int16_t& lhs, const Money& rhs);
	friend PECUNIA_EXPORT Money operator*(const std::int32_t& lhs, const Money& rhs);
	friend PECUNIA_EXPORT Money operator*(const std::int64_t& lhs, const Money& rhs);
	friend PECUNIA_EXPORT Money operator*(const FloatingPointBase& lhs, const Money& rhs);
	friend PECUNIA_EXPORT std::ostream& operator<<(std::ostream& stream, const Money& m);
	friend PECUNIA_EXPORT std::istream& operator>>(std::istream& stream, Money& m);
	friend PECUNIA_EXPORT void swap(Money& lhs, Money& rhs) noexcept;

	/**
	 * @brief Calculates the largest storable monetary major value.
	 */
	[[nodiscard]]
	MajorUnit maximumMajorValue() const;
	/**
	 * @brief Calculates the smallest storable monetary major value.
	 */
	[[nodiscard]]
	MajorUnit minimumMajorValue() const;
	/**
	 * @brief Calculates the largest storable monetary minor value, i.e. accounts for a currency’s
	 * minor digits and sub-minor digits.
	 */
	[[nodiscard]]
	MinorUnit maximumMinorValue() const;
	/**
	 * @brief Calculates the largest storable monetary amount value.
	 */
	[[nodiscard]]
	MajorUnit maximumAmountValue() const;
	/**
	 * @brief Calculates the smallest storable monetary amount value.
	 */
	[[nodiscard]]
	MajorUnit minimumAmountValue() const;

	/// The entire representation of the monetary amount using the currency’s minor unit with the
	/// extra sub-minor digits.
	MajorUnit amount_;
	/// The currency code the monetary amount represents.
	Currency code_;

public:
	/**
	 * @brief Creates an instance from the entire representation of the monetary amount using the
	 * currency’s major unit, minor unit, and the sub-minor digits.
	 *
	 * @exception OverflowError When the major unit is too large to be stored in a positive
	 * direction.
	 * @exception OverflowError When the major unit is too small to be stored in a negative
	 * direction.
	 * @exception OverflowError When the minor unit is too large to be stored.
	 * @exception OverflowError When the sub-minor unit is too large to be stored.
	 *
	 * @param amount The amount to be converted in the form MajorMinorSub-minor.
	 * @param code The ISO-4217 code representing the currency.
	 *
	 * @return An object that represents the given monetary amount in the specified currency.
	 */
	static Money fromAmount(const MajorUnit amount, const Currency& code);
	/**
	 * @brief Calculates the largest monetary value that can be used for a given currency.
	 *
	 * @param code The currency code to calculate the monetary value in.
	 *
	 * @return The monetary components of the maximum value.
	 */
	[[nodiscard]]
	static MoneyParts maximum(const Currency& code) noexcept;
	/**
	 * @brief Calculates the smallest monetary value that can be used for a given currency.
	 *
	 * @param code The currency code to calculate the monetary value in.
	 *
	 * @return The monetary components of the minimum value.
	 */
	[[nodiscard]]
	static MoneyParts minimum(const Currency& code) noexcept;

	/**
	 * @brief Constructor for a fully formed object that initialises the stored amount to an
	 * unknown currency at zero.
	 */
	Money();
	/**
	 * @brief Constructor for a fully formed object that initialises the stored amount to zero of a
	 * specific currency.
	 *
	 * @param code The currency code the amount represents.
	 */
	explicit Money(const Currency& code);

	// cppcheck-suppress noExplicitConstructor
	// Justification: It is safe and desirable to go from a Minor to Money type implicitly.

	/**
	 * @brief Constructor for a fully formed object that initialises based on the currency’s minor
	 * value.
	 *
	 * @param m The object from which to initialise the object.
	 */
	Money(const Minor& m); // NOLINT(*-explicit-constructor)
	/**
	 * @brief Constructor for setting an exact monetary value of a fully formed object for a
	 * currency that expects a minor value.
	 *
	 * @pre The maximum minor value must fit within the MinorUnit.
	 *
	 * @exception OverflowError When the major unit is too large to be stored in a positive
	 * direction.
	 * @exception OverflowError When the major unit is too small to be stored in a negative
	 * direction.
	 * @exception OverflowError When the minor unit is too large to be stored.
	 * @exception OverflowError When the sub-minor unit is too large to be stored.
	 * @exception InvalidArgument Thrown if the `minor` value is provided while the currency does
	 * not support minor digits.
	 *
	 * @param code The currency code the amount represents.
	 * @param major The amount of major currency to store, e.g. U.S.A. dollars.
	 * @param minor The amount of the minor currency to store, e.g. U.S.A. cents. For non-minor
	 * currencies, no minor unit will be set.
	 * @param subMinor The amount of the sub-minor currency to store, e.g. U.S.A.
	 * fractional cents.
	 */
	Money(
		const Currency& code,
		const MajorUnit& major,
		const MinorUnit& minor = 0u,
		const MinorUnit& subMinor = 0
	);
	/**
	 * @brief Constructor for setting an exact monetary value of a fully formed object for a
	 * currency that may support a minor value.
	 *
	 * @pre The maximum minor value must fit within the MinorUnit.
	 *
	 * @exception OverflowError When the major unit is too large to be stored in a positive
	 * direction.
	 * @exception OverflowError When the major unit is too small to be stored in a negative
	 * direction.
	 * @exception OverflowError When the minor unit is too large to be stored.
	 * @exception OverflowError When the sub-minor unit is too large to be stored.
	 * @exception InvalidArgument Thrown if the `minor` value is provided while the currency does
	 * not support minor digits.
	 * @exception InvalidArgument Thrown if the `minor` value is /not/ provided while the currency
	 * *does* support minor digits.
	 *
	 * @param code The currency code the amount represents.
	 * @param major The amount of major currency to store, e.g. U.S.A. dollars.
	 * @param minor The amount of the minor currency to store, e.g. U.S.A. cents.
	 * @param subMinor The amount of the sub-minor currency to store, e.g. U.S.A.
	 * fractional cents.
	 */
	Money(
		const Currency& code,
		const MajorUnit& major,
		const std::optional<MinorUnit>& minor,
		const MinorUnit& subMinor = 0
	);
	/**
	 * @brief Constructor for setting an exact monetary value of a fully formed object. This is most
	 * useful for the currencies that never have minor units.
	 *
	 * @exception OverflowError When the major unit is too large to be stored in a positive
	 * direction.
	 * @exception OverflowError When the major unit is too small to be stored in a negative
	 * direction.
	 *
	 * @param major The amount of major currency to store, e.g. Japan Yen.
	 * @param code The currency code the amount represents.
	 */
	Money(const MajorUnit& major, const Currency& code);
	/**
	 * @brief Constructor for setting an exact monetary value of a fully formed object. This is most
	 * useful for the currencies that never have minor units.
	 *
	 * @exception OverflowError When the major unit is too large to be stored in a positive
	 * direction.
	 * @exception OverflowError When the major unit is too small to be stored in a negative
	 * direction.
	 * @exception OverflowError When the sub-minor unit is too large to be stored.
	 *
	 * @param major The amount of major currency to store, e.g. Japan Yen.
	 * @param subMinor The amount of the sub-minor currency to store, e.g. Japan fractional Yen.
	 * @param code The currency code the amount represents.
	 */
	Money(const MajorUnit& major, const MinorUnit& subMinor, const Currency& code);
	/**
	 * @brief Constructor for setting a monetary value using the native floating-point type.
	 *
	 * @note Minor units in a floating-point value cannot be verified. They will always fit within
	 * the currency digits and precision specified, e.g. U.S.A. dollar value with sub-minor digits
	 * of 0 will store the value 1.555 as 1.55 when no rounder is supplied.
	 *
	 * @exception OverflowError When the major unit is too large to be stored in a positive
	 * direction.
	 * @exception OverflowError When the major unit is too small to be stored in a negative
	 * direction.
	 * @exception DomainError When the conversion results in the floating-point value dividing by
	 * zero.
	 * @exception UnderflowError When the floating-point value is too small to fit within the
	 * FloatingPoint.
	 * @exception OverflowError When the floating-point value is too large to fit within the
	 * FloatingPoint.
	 * @exception DomainError When the conversion results in the floating-point value not being
	 * valid.
	 *
	 * @param code The currency code the amount represents.
	 * @param value The floating-point value to convert.
	 * @param rounder The method for rounding the number when converting the floating-point
	 * number.
	 */
	Money(
		const FloatingPointRounderFunction& rounder,
		const Currency& code,
		const FloatingPointBase& value
	);

	Money(const Money&) = default;

	Money(Money&&) noexcept = default;
	/**
	 * @brief Conversion operator to a native floating-point type.
	 *
	 * @exception DomainError When the conversion results in the floating-point value dividing by
	 * zero.
	 * @exception UnderflowError When the floating-point value is too small to fit within the
	 * FloatingPoint.
	 * @exception OverflowError When the floating-point value is too large to fit within the
	 * FloatingPoint.
	 * @exception DomainError When the conversion results in the floating-point value not being
	 * valid.
	 */
	explicit operator FloatingPointBase() const;

	~Money() = default;
	/**
	 * @brief Negation operator without any loss of precision during the conversion.
	 *
	 * @return The negated value of the current object.
	 */
	[[nodiscard]]
	Money operator-() const;
	/**
	 * @brief The additive operator. When the supplied monetary object is of a different currency,
	 * the amount is converted into the current object’s currency using the
	 * pecunia::converter function.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param other The additional value to add to the current object’s value.
	 *
	 * @return A new object containing the summation of the two values.
	 */
	[[nodiscard]]
	Money operator+(const Money& other) const;
	/**
	 * @brief The subtractive operator. When the supplied monetary object is of a different
	 * currency, the amount is converted into the current object’s currency using the
	 * pecunia::converter function.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param other The additional value to subtract from the current object’s value.
	 *
	 * @return A new object containing the difference of the two values.
	 */
	[[nodiscard]]
	Money operator-(const Money& other) const;

	Money operator*(const Money&) const = delete;
	/**
	 * @brief The multiplicative operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The signed 8-bit integer value to multiply to the current object’s value.
	 *
	 * @return A new object containing the product of the two values.
	 */
	[[nodiscard]]
	Money operator*(const std::int8_t& value) const;
	/**
	 * @brief The multiplicative operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The signed 16-bit integer value to multiply to the current object’s value.
	 *
	 * @return A new object containing the product of the two values.
	 */
	[[nodiscard]]
	Money operator*(const std::int16_t& value) const;
	/**
	 * @brief The multiplicative operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The signed 32-bit integer value to multiply to the current object’s value.
	 *
	 * @return A new object containing the product of the two values.
	 */
	[[nodiscard]]
	Money operator*(const std::int32_t& value) const;
	/**
	 * @brief The multiplicative operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The signed 64-bit integer value to multiply to the current object’s value.
	 *
	 * @return A new object containing the product of the two values.
	 */
	[[nodiscard]]
	Money operator*(const std::int64_t& value) const;
	/**
	 * @brief The multiplicative operator without any loss of precision. This only supports as many
	 * digits as minor digits.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional value to multiply to the current object’s value.
	 *
	 * @return A new object containing the product of the two values.
	 */
	[[nodiscard]]
	Money operator*(const FloatingPointBase& value) const;
	/**
	 * @brief The division operator for calculating ratios.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 *
	 * @param other The value to divide by the current object’s value.
	 *
	 * @return The result of the division between the two, a ratio.
	 */
	[[nodiscard]]
	FloatingPointBase operator/(const Money& other) const;
	/**
	 * @brief Division operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional signed 8-bit integer value to divide by the current object’s
	 * value.
	 *
	 * @return A new object containing the division of the two values.
	 */
	[[nodiscard]]
	Money operator/(const std::int8_t& value) const;
	/**
	 * @brief Division operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional signed 16-bit integer value to divide by the current object’s
	 * value.
	 *
	 * @return A new object containing the division of the two values.
	 */
	[[nodiscard]]
	Money operator/(const std::int16_t& value) const;
	/**
	 * @brief Division operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional signed 32-bit integer value to divide by the current object’s
	 * value.
	 *
	 * @return A new object containing the division of the two values.
	 */
	[[nodiscard]]
	Money operator/(const std::int32_t& value) const;
	/**
	 * @brief Division operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional signed 64-bit integer value to divide by the current object’s
	 * value.
	 *
	 * @return A new object containing the division of the two values.
	 */
	[[nodiscard]]
	Money operator/(const std::int64_t& value) const;
	/**
	 * @brief Division operator without any loss of precision. This only supports as many digits as
	 * minor digits.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional value to divide by the current object’s value.
	 *
	 * @return A new object containing the division of the two values.
	 */
	[[nodiscard]]
	Money operator/(const FloatingPointBase& value) const;

	Money operator%(const Money&) const = delete;
	/**
	 * @brief Modulus operator performed upon /only/ the major units of the object’s current value.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 *
	 * @param value The additional value to perform a mod on the current object’s value.
	 *
	 * @return The value of the modulus operation from the two values.
	 */
	[[nodiscard]]
	MajorUnit operator%(const MajorUnit& value) const;
	/**
	 * @brief The assignment Operator with a copy. When the supplied monetary object is of a
	 * different currency, the amount is converted into the current object’s currency using the
	 * pecunia::converter function.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception OverflowError When the result of the operation would be too small to store.
	 *
	 * @param other The object whose amount is to be assigned into the current object.
	 *
	 * @post The current object’s amount is changed to the supplied one. The original object’s
	 * currency does not change.
	 *
	 * @return The current object with the new amount.
	 */
	Money& operator=(const Money& other);
	/**
	 * @brief Less-than Relational Operator. When the supplied monetary object is of a different
	 * currency, the amount is converted into the current object’s currency using the
	 * pecunia::converter function.
	 *
	 * @param other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<(const Money& other) const;
	/**
	 * @brief Less-than Relational Operator for an 8-bit signed integer. The supplied value is
	 * expressed in terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<(const std::int8_t& value) const;
	/**
	 * @brief Less-than Relational Operator for a 16-bit signed integer. The supplied value is
	 * expressed in terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<(const std::int16_t& value) const;
	/**
	 * @brief Less-than Relational Operator for a 32-bit signed integer. The supplied value is
	 * expressed in terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<(const std::int32_t& value) const;
	/**
	 * @brief Less-than Relational Operator for a 64-bit signed integer. The supplied value is
	 * expressed in terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<(const std::int64_t& value) const;
	/**
	 * @brief Less-than-equal Relational Operator. When the supplied monetary object is of a
	 * different currency, the amount is converted into the current object’s currency using the
	 * pecunia::converter function.
	 *
	 * @param other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than or equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<=(const Money& other) const;
	/**
	 * @brief Less-than-equal Relational Operator for an 8-bit signed integer. The supplied value is
	 * expressed in terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than or equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<=(const std::int8_t& value) const;
	/**
	 * @brief Less-than-equal Relational Operator for a 16-bit signed integer. The supplied value
	 * is expressed in terms of the current object’s issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than or equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<=(const std::int16_t& value) const;
	/**
	 * @brief Less-than-equal Relational Operator for a 32-bit signed integer. The supplied value
	 * is expressed in terms of the current object’s issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than or equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<=(const std::int32_t& value) const;
	/**
	 * @brief Less-than-equal Relational Operator for a 64-bit signed integer. The supplied value
	 * is expressed in terms of the current object’s issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is less than or equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator<=(const std::int64_t& value) const;
	/**
	 * @brief Greater-than Relational Operator. When the supplied monetary object is of a different
	 * currency, the amount is converted into the current object’s currency using the
	 * pecunia::converter function.
	 *
	 * @param other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator>(const Money& other) const;
	/**
	 * @brief Greater-than Relational Operator for an 8-bit signed integer. The supplied value is
	 * expressed in terms of the current object’s issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator>(const std::int8_t& value) const;
	/**
	 * @brief Greater-than Relational Operator for a 16-bit signed integer. The supplied value is
	 * expressed in terms of the current object’s issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator>(const std::int16_t& value) const;
	/**
	 * @brief Greater-than Relational Operator for a 32-bit signed integer. The supplied value is
	 * expressed in terms of the current object’s issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator>(const std::int32_t& value) const;
	/**
	 * @brief Greater-than Relational Operator for a 64-bit signed integer. The supplied value is
	 * expressed in terms of the current object’s issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator>(const std::int64_t& value) const;
	/**
	 * @brief Greater-than-equal Relational Operator. When the supplied monetary object is of a
	 * different currency, the amount is converted into the current object’s currency using the
	 * pecunia::converter function.
	 *
	 * @param other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than or equal to the supplied
	 * object, false else-wise.
	 */
	[[nodiscard]]
	bool operator>=(const Money& other) const;
	/**
	 * @brief Greater-than-equal Relational Operator for an 8-bit signed integer. The supplied value
	 * is expressed in terms of the current object’s issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than or equal to the supplied
	 * object, false else-wise.
	 */
	[[nodiscard]]
	bool operator>=(const std::int8_t& value) const;
	/**
	 * @brief Greater-than-equal Relational Operator for a 16-bit signed integer. The supplied
	 * value is expressed in terms of the current object’s issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than or equal to the supplied
	 * object, false else-wise.
	 */
	[[nodiscard]]
	bool operator>=(const std::int16_t& value) const;
	/**
	 * @brief Greater-than-equal Relational Operator for a 32-bit signed integer. The supplied
	 * value is expressed in terms of the current object’s issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than or equal to the supplied
	 * object, false else-wise.
	 */
	[[nodiscard]]
	bool operator>=(const std::int32_t& value) const;
	/**
	 * @brief Greater-than-equal Relational Operator for a 64-bit signed integer. The supplied
	 * value is expressed in terms of the current object’s issued currency, in the
	 * form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is greater than or equal to the supplied
	 * object, false else-wise.
	 */
	[[nodiscard]]
	bool operator>=(const std::int64_t& value) const;
	/**
	 * @brief Equality Operator. When the supplied monetary object is of a different currency,
	 * the amount is converted into the current object’s currency using the
	 * pecunia::converter function.
	 *
	 * @param other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator==(const Money& other) const;
	/**
	 * @brief Equality Operator for an 8-bit signed integer. The supplied value is expressed in
	 * terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is equal to the supplied value within the
	 * expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator==(const std::int8_t& value) const;
	/**
	 * @brief Equality Operator for a 16-bit signed integer. The supplied value is expressed in
	 * terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is equal to the supplied value within the
	 * expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator==(const std::int16_t& value) const;
	/**
	 * @brief Equality Operator for a 32-bit signed integer. The supplied value is expressed in
	 * terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is equal to the supplied value within the
	 * expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator==(const std::int32_t& value) const;
	/**
	 * @brief Equality Operator for a 64-bit signed integer. The supplied value is expressed in
	 * terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is equal to the supplied value within the
	 * expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator==(const std::int64_t& value) const;
	/**
	 * @brief Inequality Operator. When the supplied monetary object is of a different currency,
	 * the amount is converted into the current object’s currency using the
	 * pecunia::converter function.
	 *
	 * @param other The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is not equal to the supplied object,
	 * false else-wise.
	 */
	[[nodiscard]]
	bool operator!=(const Money& other) const;
	/**
	 * @brief Inequality Operator for an 8-bit signed integer. The supplied value is expressed in
	 * terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is not equal to the supplied value, within
	 * the expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator!=(const std::int8_t& value) const;
	/**
	 * @brief Inequality Operator for a 16-bit signed integer. The supplied value is expressed in
	 * terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is not equal to the supplied value, within
	 * the expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator!=(const std::int16_t& value) const;
	/**
	 * @brief Inequality Operator for a 32-bit signed integer. The supplied value is expressed in
	 * terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is not equal to the supplied value, within
	 * the expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator!=(const std::int32_t& value) const;
	/**
	 * @brief Inequality Operator for a 64-bit signed integer. The supplied value is expressed in
	 * terms of the current object’s issued currency, in the form MajorMinorSub-minor.
	 *
	 * @param value The other object whose value is to be compared against.
	 *
	 * @return A value of true with the current object is not equal to the supplied value, within
	 * the expected level of precision, false else-wise.
	 */
	[[nodiscard]]
	bool operator!=(const std::int64_t& value) const;
	/**
	 * @brief Additive assignment operator. When the supplied monetary object is of a different
	 * currency, the amount is converted into the current object’s currency using the
	 * pecunia::converter function.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param other The additional value to add to the current object.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the summation of the two values.
	 */
	Money& operator+=(const Money& other);
	/**
	 * @brief Subtractive assignment operator. When the supplied monetary object is of a different
	 * currency, the amount is converted into the current object’s currency using the
	 * pecunia::converter function.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param other The additional value to subtract from the current object.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the difference of the two values.
	 */
	Money& operator-=(const Money& other);
	/**
	 * @brief Multiplicative assignment operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional signed 8-bit integer value to multiply to the current
	 * object.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the product of the two values.
	 */
	Money& operator*=(const std::int8_t& value);
	/**
	 * @brief Multiplicative assignment operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional signed 16-bit integer value to multiply to the current
	 * object.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the product of the two values.
	 */
	Money& operator*=(const std::int16_t& value);
	/**
	 * @brief Multiplicative assignment operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional signed 32-bit integer value to multiply to the current
	 * object.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the product of the two values.
	 */
	Money& operator*=(const std::int32_t& value);
	/**
	 * @brief Multiplicative assignment operator without any loss of precision.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional signed 64-bit integer value to multiply to the current
	 * object.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the product of the two values.
	 */
	Money& operator*=(const std::int64_t& value);
	/**
	 * @brief Multiplicative assignment operator without any loss of precision. This only supports
	 * as many digits as minor digits.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional value to multiply to the current object.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the product of the two values.
	 */
	Money& operator*=(const FloatingPointBase& value);
	/**
	 * @brief Division assignment operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The signed 8-bit integer additional value to divide by the current object’s
	 * value.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the division of the two values.
	 */
	Money& operator/=(const std::int8_t& value);
	/**
	 * @brief Division assignment operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The signed 16-bit integer additional value to divide by the current object’s
	 * value.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the division of the two values.
	 */
	Money& operator/=(const std::int16_t& value);
	/**
	 * @brief Division assignment operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The signed 32-bit integer additional value to divide by the current object’s
	 * value.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the division of the two values.
	 */
	Money& operator/=(const std::int32_t& value);
	/**
	 * @brief Division assignment operator without any loss of precision.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The signed 64-bit integer additional value to divide by the current object’s
	 * value.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the division of the two values.
	 */
	Money& operator/=(const std::int64_t& value);
	/**
	 * @brief Division assignment operator without any loss of precision. This only supports as many
	 * digits as minor digits.
	 *
	 * @exception DomainError When the result of the operation would result in a division by zero.
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception UnderflowError When the result of the operation would be too small to store.
	 *
	 * @param value The additional value to divide by the current object’s value.
	 *
	 * @post The current object’s amount is changed using the mathematical operation. The current
	 * object’s currency does not change.
	 *
	 * @return The current object containing the division of the two values.
	 */
	Money& operator/=(const FloatingPointBase& value);
	/**
	 * @brief The entire representation of the monetary amount using the currency’s major unit,
	 * minor unit, and the sub-minor digits, in the form MajorMinorSub-minor.
	 */
	[[nodiscard]]
	const MajorUnit& amount() const noexcept;
	/**
	 * @brief Accessor to the major monetary value, e.g. only the U.S.A. dollars.
	 */
	[[nodiscard]]
	MajorUnit major() const;
	/**
	 * @brief Accessor to the minor and sub-minor monetary value, e.g. the U.S.A. cents and
	 * fractional cents.
	 */
	[[nodiscard]]
	MinorUnit minorWithSub() const;
	/**
	 * @brief Accessor to the minor monetary value, e.g. the U.S.A. cents, truncating the fractional
	 * cents.
	 */
	[[nodiscard]]
	std::optional<MinorUnit> minorOnly() const;
	/**
	 * @brief Accessor to the sub-minor monetary value, e.g. only the U.S.A. fractional cents.
	 */
	[[nodiscard]]
	MinorUnit subMinor() const;
	/**
	 * @brief Accessor to the code the currency amount represents.
	 */
	[[nodiscard]]
	const Currency& code() const noexcept;
	/**
	 * @brief Converts the monetary value into a native floating-point type using a supplied
	 * rounding method.
	 *
	 * @param rounder The rounding function to apply to the monetary value.
	 * @param digits  The number of digits after the major unit.
	 *
	 * @exception DomainError When the conversion results in the floating-point value dividing by
	 * zero.
	 * @exception UnderflowError When the floating-point value is too small to fit within the
	 * FloatingPoint.
	 * @exception OverflowError When the floating-point value is too large to fit within the
	 * FloatingPoint.
	 * @exception DomainError When the conversion results in the floating-point value not being
	 * valid.
	 */
	[[nodiscard]]
	FloatingPointBase toFloatingPoint(
		const RounderFunction& rounder,
		const std::uint8_t digits = defaultRoundingDigits
	) const;
	/**
	 * @brief Performs a rounding operation upon the monetary value.
	 *
	 * @exception OverflowError When the result of the operation would be too large to store.
	 * @exception OverflowError When the result of the operation would be too small to store.
	 * @exception DomainError When the conversion results in the floating-point value dividing by
	 * zero.
	 * @exception UnderflowError When the floating-point value is too small to fit within the
	 * FloatingPoint.
	 * @exception OverflowError When the floating-point value is too large to fit within the
	 * FloatingPoint.
	 * @exception DomainError When the conversion results in the floating-point value not being
	 * valid.
	 *
	 * @param rounder The rounding function to apply to the monetary value.
	 * @param digits The number of digits after the major unit to round to.
	 *
	 * @return A copy of the current object containing the rounded value.
	 */
	[[nodiscard]]
	Money round(
		const RounderFunction& rounder,
		const std::uint8_t digits = defaultRoundingDigits
	) const;
	/**
	 * @brief Resets the monetary value to zero, without changing the currency.
	 */
	void zeroOut();
	/**
	 * @brief Determines if the monetary amount consists of only minor units, i.e. only change.
	 *
	 * @return Gives true when the amount is exclusively between negative one major unit and one
	 * major unit, false otherwise.
	 */
	[[nodiscard]]
	bool hasOnlyMinor() const noexcept;
	/**
	 * @brief Assigns the value and currency from the provided object to the current instance.
	 *
	 * @param other The object whose value and currency will be assigned.
	 *
	 * @post The current instance will have the same value and currency as the provided object.
	 *
	 * @return The current instance after assignment.
	 */
	Money& assign(const Money& other) noexcept;
	/**
	 * @brief Assigns the value and currency from the provided object to the current instance.
	 *
	 * @param other The object whose value and currency will be assigned.
	 *
	 * @post The current instance will have the same value and currency as the provided object.
	 * @post The `other` instance will have its internal properties reset or become invalid.
	 *
	 * @return The current instance after assignment.
	 */
	Money& assign(Money&& other) noexcept;
	/**
	 * @brief Resets the object to its default state.
	 *
	 * @post The object is reset to a default-constructed state.
	 */
	void clear() noexcept;
};

/**
 * @brief The multiplicative operator without any loss of precision.
 *
 * @exception OverflowError When the result of the operation would be too large to store.
 * @exception UnderflowError When the result of the operation would be too small to store.
 *
 * @param lhs The signed 8-bit integer amount by which to multiply the supplied object’s
 * value.
 * @param rhs The object that is to be multiplied.
 *
 * @return A new object containing the product of the two values.
 */
[[nodiscard]]
PECUNIA_EXPORT Money operator*(const std::int8_t& lhs, const Money& rhs);
/**
 * @brief The multiplicative operator without any loss of precision.
 *
 * @exception OverflowError When the result of the operation would be too large to store.
 * @exception UnderflowError When the result of the operation would be too small to store.
 *
 * @param lhs The signed 16-bit integer amount by which to multiply the supplied object’s
 * value.
 * @param rhs The object that is to be multiplied.
 *
 * @return A new object containing the product of the two values.
 */
[[nodiscard]]
PECUNIA_EXPORT Money operator*(const std::int16_t& lhs, const Money& rhs);
/**
 * @brief The multiplicative operator without any loss of precision.
 *
 * @exception OverflowError When the result of the operation would be too large to store.
 * @exception UnderflowError When the result of the operation would be too small to store.
 *
 * @param lhs The signed 32-bit integer amount by which to multiply the supplied object’s
 * value.
 * @param rhs The object that is to be multiplied.
 *
 * @return A new object containing the product of the two values.
 */
[[nodiscard]]
PECUNIA_EXPORT Money operator*(const std::int32_t& lhs, const Money& rhs);
/**
 * @brief The multiplicative operator without any loss of precision.
 *
 * @exception OverflowError When the result of the operation would be too large to store.
 * @exception UnderflowError When the result of the operation would be too small to store.
 *
 * @param lhs The signed 64-bit integer amount by which to multiply the supplied object’s
 * value.
 * @param rhs The object that is to be multiplied.
 *
 * @return A new object containing the product of the two values.
 */
[[nodiscard]]
PECUNIA_EXPORT Money operator*(const std::int64_t& lhs, const Money& rhs);
/**
 * @brief The multiplicative operator without any loss of precision. This only supports as many
 * digits as minor digits.
 *
 * @exception OverflowError When the result of the operation would be too large to store.
 * @exception UnderflowError When the result of the operation would be too small to store.
 *
 * @param lhs The amount by which to multiply the supplied object’s value.
 * @param rhs The object that is to be multiplied.
 *
 * @return A new object containing the product of the two values.
 */
[[nodiscard]]
PECUNIA_EXPORT Money operator*(const FloatingPointBase& lhs, const Money& rhs);
/**
 * @brief Stream Insertion Operator. By default, the form will be in the locale’s monetary format
 * with no spaces, the ISO code, and trailing the value, e.g. a U.S.A. locale with value of one
 * dollar and twenty-three cents would appear as 1.23USD.
 *
 * @exception RuntimeError When the stream fails to be written.
 *
 * @param stream The destination source for the contents of the supplied money object.
 * @param m The object whose values are placed into the destination source.
 *
 * @return The modified source stream object filled with the contents of the money object.
 */
PECUNIA_EXPORT std::ostream& operator<<(std::ostream& stream, const Money& m);
/**
 * @brief Stream Extraction Operator. The monetary amount will be read in using the locale’s setting
 * for reading in money. The currency code is required and must be in all upper-case letters.
 *
 * @pre The locale of the supplied stream must be set.
 *
 * @exception RuntimeError When the stream fails to be read.
 * @exception InvalidArgument When the currency is incorrectly formatted.
 *
 * @param stream The extraction source for the contents of the supplied money object.
 * @param m The object whose values are to be filled by the contents of the stream.
 *
 * @return The modified source stream object emptied of a money entry.
 */
PECUNIA_EXPORT std::istream& operator>>(std::istream& stream, Money& m);
/**
 * @brief Converts the monetary representation into a string from. The format of the string
 * representation is the amount followed by a space, and the currency code.
 *
 * @param m The object whose values are to be converted into a string form.
 *
 * @return The string representation of the monetary amount.
 */
[[nodiscard]]
PECUNIA_EXPORT std::string toStdString(const Money& m);
/**
 * @brief Exchanges two monetary objects without any conversion operation performed.
 *
 * @param lhs The object to swap to the right.
 * @param rhs The object to swap to the left.
 */
PECUNIA_EXPORT void swap(Money& lhs, Money& rhs) noexcept;

}

#endif
