/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024, 2025            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_MONEY_PARTS_HPP_
#define PECUNIA_MONEY_PARTS_HPP_

#include <cstdint>
#include <optional>

#ifdef __WIN32
#	pragma STDC FENV_ACCESS ON
#endif

#include <pecunia/pecunia_export.h>
#include <pecunia/Sizes.hpp>


namespace pecunia
{

enum class Currency : std::uint16_t;
class FloatingPoint;
class Money;

/**
 * @brief Represents the components of a monetary value separated into major, minor, sub-minor units
 * and its associated currency.
 *
 * It is designed to facilitate the manipulation and representation of complex monetary values
 * by providing segmented access to the major, minor, and sub-minor components, along with the
 * currency. It is primarily used for scenarios requiring precise control and interpretation of
 * monetary values, formatting, etc.
 */
class MoneyParts
{
	// Justification: Used for performance reasons.
	friend class Money;

	MoneyParts(
		const MajorUnit full,
		const MajorUnit major,
		const MinorUnit minorWithSub,
		const std::optional<MinorUnit> minorOnly,
		const MinorUnit subMinor,
		const Currency code
	);

	MajorUnit full_;
	MajorUnit major_;
	MinorUnit minorWithSub_;
	std::optional<MinorUnit> minorOnly_;
	MinorUnit subMinor_;
	Currency code_;

public:
	/**
	 * @brief Constructs an empty object with default initialised values.
	 *
	 * This constructor is primarily used to create an instance with zero values for all its
	 * member properties, and a default currency code of XXX, usually as a placeholder or initial
	 * state.
	 */
	PECUNIA_EXPORT MoneyParts();
	/**
	 * @brief Full initialisation constructor that fully initialises the object.
	 *
	 * @exception InvalidArgument When the monetary value does not support minor units and minor
	 * units are supplied.
	 *
	 * @param code The currency of the monetary value.
	 * @param major The major monetary value.
	 * @param minorOnly The minor monetary value, if the currency supports it.
	 * @param subMinor The sub-minor monetary value.
	 */
	PECUNIA_EXPORT MoneyParts(
		const Currency code,
		const MajorUnit major,
		const std::optional<MinorUnit> minorOnly,
		const MinorUnit subMinor = 0
	);
	/**
	 * @brief Retrieves the full monetary value represented in its major unit including all minor
	 * components.
	 *
	 * This method provides access to the total monetary value as represented in the major unit of
	 * the currency, incorporating all minor and sub-minor units.
	 *
	 * @return The full monetary value in the form MajorMinorSub-minor.
	 */
	[[nodiscard]]
	PECUNIA_EXPORT const MajorUnit& full() const noexcept;
	/**
	 * @brief Retrieves the major unit part of the monetary value.
	 *
	 * This method is used to extract only the integral, major unit component of a monetary
	 * value, excluding any minor or sub-minor parts.
	 *
	 * @return The major unit value of this monetary object.
	 */
	[[nodiscard]]
	PECUNIA_EXPORT const MajorUnit& major() const noexcept;
	/**
	 * @brief Retrieves the combined value of the minor and sub-minor units.
	 *
	 * This method provides access to the aggregated minor and sub-minor components of a monetary
	 * value.
	 *
	 * @return The combined value of minor and sub-minor units in the form MinorSub-minor.
	 */
	[[nodiscard]]
	PECUNIA_EXPORT const MinorUnit& minorWithSub() const noexcept;
	/**
	 * @brief Retrieves the minor-only part of a monetary value, excluding other components.
	 *
	 * @return The minor unit of the monetary value, if it exists, null when it does not.
	 */
	[[nodiscard]]
	PECUNIA_EXPORT const std::optional<MinorUnit>& minorOnly() const noexcept;
	/**
	 * @brief Retrieves the sub-minor unit associated with the monetary value. This provides
	 * fine-grained detail of the monetary amount, particularly when precision to the smallest unit
	 * (e.g., fractional cents) is required.
	 *
	 * @return The sub-minor unit of the monetary value.
	 */
	[[nodiscard]]
	PECUNIA_EXPORT const MinorUnit& subMinor() const noexcept;
	/**
	 * @brief Retrieves the currency code of the monetary value.
	 *
	 * @return The currency code of the monetary value, e.g. USD, GBP, EUR, etc.
	 */
	[[nodiscard]]
	PECUNIA_EXPORT const Currency& code() const noexcept;
};

/**
 * @brief Compares two objects to determine if the first is less than the second, considering both
 * their currency codes and monetary values.
 *
 * This function is used to perform comparisons between monetary values, ensuring accurate results
 * by factoring in both the currency codes, and the actual monetary amounts. It is useful in
 * sorting or scenarios requiring strict order of monetary parts.
 *
 * @param lhs The first object to compare.
 * @param rhs The second object to compare.
 *
 * @return True if the lhs object is considered less than the rhs object; otherwise, false.
 * Consideration is in the order currency then monetary value.
 */
[[nodiscard]]
PECUNIA_EXPORT bool operator<(const MoneyParts& lhs, const MoneyParts& rhs);
/**
 * @brief General equality operator.
 *
 * @param lhs The object that is expected to be compared.
 * @param rhs The object that is to compare for equality.
 *
 * @return True when the currency and monetary value are the same, false otherwise.
 */
[[nodiscard]]
PECUNIA_EXPORT bool operator==(const MoneyParts& lhs, const MoneyParts& rhs);
/**
 * @brief General inequality operator.
 *
 * @param lhs The object that is expected to be compared.
 * @param rhs The object that is to compare for inequality.
 *
 * @return False when the currency and monetary value are the same, true otherwise.
 */
[[nodiscard]]
PECUNIA_EXPORT bool operator!=(const MoneyParts& lhs, const MoneyParts& rhs);

}

#endif
