/*******************************************************************************
***  This file is part of Pecunia.                                           ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2023, 2024, 2025            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef PECUNIA_SETUP_HPP_
#define PECUNIA_SETUP_HPP_

#include <string>

#include <pecunia/pecunia_export.h>

#include "Conversion.hpp"


namespace pecunia
{

/**
 * @brief Sets up the behaviour for the library, ensuring a suitable default environment is
 * established for monetary computations.
 *
 * @exception RuntimeError Thrown if floating-point exception flags cannot be cleared.
 *
 * @post The currency conversion function only works when the currencies are the same.
 * @post The locale is set to the default configured value.
 * @post All floating-point exception flags are cleared.
 */
PECUNIA_EXPORT void setUpLibrary();
/**
 * @brief Sets up currency-related behaviours for the library. This ensures the proper environment
 * is established for currency operations.
 *
 * @pre The converterFunctor must point to a valid function and must not be null.
 *
 * @exception RuntimeError Thrown if floating-point exception flags cannot be cleared.
 *
 * @param converterFunctor The function used for currency conversion.
 *
 * @post The currency conversion function is successfully established.
 * @post All floating-point exception flags are cleared.
 */
PECUNIA_EXPORT void setUpLibrary(const ConversionFunction& converterFunctor);
/**
 * @brief Sets up currency-related behaviours for the library. This ensures the proper environment
 * is established for currency operations.
 *
 * @pre The locale should represent a valid locale format supported by the system.
 *
 * @exception RuntimeError Thrown if the locale cannot be set to the given input.
 * @exception RuntimeError Thrown if floating-point exception flags cannot be cleared.
 *
 * @param localeName The locale to set for currency localisation.
 *
 * @post The currency conversion function only works when the currencies are the same.
 * @post The locale is set to the specified value.
 * @post All floating-point exception flags are cleared.
 */
PECUNIA_EXPORT void setUpLibrary(const std::string& localeName);
/**
 * @brief Sets up currency-related behaviours for the library. This ensures the proper environment
 * is established for currency operations.
 *
 * @pre The converterFunctor must point to a valid function and must not be null.
 * @pre The locale should represent a valid locale format supported by the system.
 *
 * @exception RuntimeError Thrown if the locale cannot be set to the given input.
 * @exception RuntimeError Thrown if floating-point exception flags cannot be cleared.
 *
 * @param converterFunctor The function used for currency conversion.
 * @param localeName The locale to set for currency localisation.
 *
 * @post The currency conversion function is successfully established.
 * @post The locale is set to the specified value.
 * @post All floating-point exception flags are cleared.
 */
PECUNIA_EXPORT void setUpLibrary(
	const ConversionFunction& converterFunctor,
	const std::string& localeName
);

}

#endif
